/*
    Copyright (C) 2004 Jesse Chappell
    Copyright (C) 2003 Jack O'Quin
    Copyright (C) 2001 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sndfile.h>
#include <signal.h>
#include <pthread.h>
#include <getopt.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>
#include <jack/transport.h>


#include <ffmpeg/avcodec.h>
#include "memops.h"

#include "version.h"

typedef struct _thread_info {
	pthread_t thread_id;
	AVCodecContext *av_context;
	int av_frame_size;
	FILE *sf;
	jack_nframes_t duration;
	jack_nframes_t rb_size;
	jack_client_t *client;
	dither_state_t *dither_state;
	unsigned int channels;
	int bitdepth;
	int bitrate;
	unsigned int sample_rate;
	char *path;
	int use_transport;
	int quiet;
	volatile int can_capture;
	volatile int can_process;
	volatile int do_write;
	volatile int status;
} thread_info_t;



/* JACK data */
unsigned int nports;
jack_port_t **ports;
jack_default_audio_sample_t **in;
jack_nframes_t nframes;
const size_t sample_size = sizeof(jack_default_audio_sample_t);

/* Synchronization between process thread and work thread. */
#define DEFAULT_RB_SIZE 16384		/* ringbuffer size in frames */
jack_ringbuffer_t *rb;
pthread_mutex_t work_thread_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  data_ready = PTHREAD_COND_INITIALIZER;
int shutdown = 0;
long overruns = 0;


static void* watchdog_thread(void* arg)
{
  sigset_t signalset;
  int signalno;
  int exiting = 0;
  
  while (!exiting)
  {
	  sigemptyset(&signalset);
	  
	  /* handle the following signals explicitly */
	  sigaddset(&signalset, SIGTERM);
	  sigaddset(&signalset, SIGINT);
	  sigaddset(&signalset, SIGHUP);
	  sigaddset(&signalset, SIGPIPE);
	  
	  /* block until a signal received */
	  sigwait(&signalset, &signalno);
	  
	  /* fprintf (stderr, "recieved signal %d\n", signalno); */
	  
	  if (signalno == SIGHUP) {
		  exiting = 1;
	  }
	  else {
		  exiting = 1;
	  }
  }

  
  shutdown = 1;
  
  /* to keep the compilers happy; never actually executed */
  return(0);
}



/**
 * Sets up a signal mask with sigaction() that blocks 
 * all common signals, and then launces an watchdog
 * thread that waits on the blocked signals using
 * sigwait().
 */
static void setupSignals()
{
  pthread_t watchdog;
  int res;

  /* man pthread_sigmask:
   *  "...signal actions and signal handlers, as set with
   *   sigaction(2), are shared between all threads"
   */

  struct sigaction blockaction;
  blockaction.sa_handler = SIG_IGN;
  sigemptyset(&blockaction.sa_mask);
  blockaction.sa_flags = 0;

  /* ignore the following signals */
  sigaction(SIGTERM, &blockaction, 0);
  sigaction(SIGINT, &blockaction, 0);
  sigaction(SIGHUP, &blockaction, 0);
  sigaction(SIGPIPE, &blockaction, 0);
  sigaction(SIGABRT, &blockaction, 0);

  res = pthread_create(&watchdog, 
			   NULL, 
			   watchdog_thread, 
			   NULL);
  if (res != 0) {
     fprintf(stderr, "ac3jack: Warning! Unable to create watchdog thread.\n");
  }
}



void *
work_thread (void *arg)
{
	thread_info_t *info = (thread_info_t *) arg;
	static jack_nframes_t total_captured = 0;
	jack_nframes_t samples_per_frame = info->channels;
	size_t bytes_per_frame = samples_per_frame * sample_size;
	size_t frames_at_once = 512;
	size_t framebuf_size = bytes_per_frame * frames_at_once;
	char *framebuf = malloc (frames_at_once * bytes_per_frame);
	
	/* 16bit input to libavcodec */
	size_t ac3_bytes_per_frame = info->av_context->channels * sizeof(short);
	size_t ac3_framebuf_size = frames_at_once * ac3_bytes_per_frame;
	char *ac3_framebuf =  malloc(ac3_framebuf_size); 
	size_t one_ac3_frame_bytes = ac3_bytes_per_frame * info->av_frame_size;
	
	size_t outbuf_size = 16384;
	char * outbuf = malloc(outbuf_size);
	size_t outsize;
	int writing = 1;
	
	jack_ringbuffer_t *ac3_rb = jack_ringbuffer_create (65536);
	
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	pthread_mutex_lock (&work_thread_lock);

	info->status = 0;

	while (1) {

		while (info->can_capture &&
		       (jack_ringbuffer_read_space (rb) >= framebuf_size)) {

			jack_ringbuffer_read (rb, framebuf, framebuf_size);

			/* convert to 16 bit
			   TODO: choice of dither */
			sample_move_d16_sS ((char *) ac3_framebuf, (jack_default_audio_sample_t *) framebuf,
					    samples_per_frame * frames_at_once, sizeof(short), info->dither_state);
			
			/* printf ("got %d bytes\n", framebuf_size); */
			
			jack_ringbuffer_write (ac3_rb, ac3_framebuf, ac3_framebuf_size); 
			
			while (jack_ringbuffer_read_space (ac3_rb) >= one_ac3_frame_bytes)
			{
				jack_ringbuffer_read (ac3_rb, ac3_framebuf, one_ac3_frame_bytes);

				if (!info->do_write) {
					/* we keep the frame blocking in sync, just don't encode or write */
					if (writing && !info->quiet) {
						fprintf(stderr, "transport stopped, output disabled...\n");
						writing = 0;
					}
					continue;
				}
				else {
					if (!writing && !info->quiet) {
						fprintf(stderr, "transport started, output enabled...\n");
						writing = 1;
					}
				}
				
				outsize = avcodec_encode_audio(info->av_context, outbuf, outbuf_size, (short *) ac3_framebuf);

				/* TODO: write to raw spdif or file */
				if (info->sf) {
					fwrite (outbuf, outsize, 1, info->sf);
				}

				/*printf ("one ac3 framebytes=%d  outsize returned %d\n", one_ac3_frame_bytes, outsize); */
				
			}

			total_captured += frames_at_once;
 			if (info->duration && ++total_captured >= info->duration) { 
				if (!info->quiet) {
					fprintf (stderr, "duration expired, done\n");
				}
 				goto done; 
 			} 
			if (shutdown) {
				if (!info->quiet) {
					fprintf (stderr, "ac3jack interrupted: shutting down\n");
				}
				goto done;
			}
		}

		/* wait until process() signals more data */
		pthread_cond_wait (&data_ready, &work_thread_lock);
	}

 done:
	info->can_capture = 0;
	pthread_mutex_unlock (&work_thread_lock);

	/* XXX: let these leak, for some reason trying to
	   free them causes segfaults */
/* 	free (framebuf); */
/* 	free (ac3_framebuf); */
/* 	free (outbuf); */
/* 	jack_ringbuffer_free (ac3_rb); */

	
	return 0;
}
	
int
process (jack_nframes_t nframes, void *arg)
{
	int chn;
	size_t i;
	thread_info_t *info = (thread_info_t *) arg;
	jack_transport_state_t tstate;
	
	/* Do nothing until we're ready to begin. */
	if ((!info->can_process) || (!info->can_capture))
		return 0;

	if (info->use_transport) {
		tstate = jack_transport_query (info->client, NULL);
		info->do_write = (tstate == JackTransportRolling) ? 1 : 0;
	}
	
	for (chn = 0; chn < nports; chn++)
		in[chn] = jack_port_get_buffer (ports[chn], nframes);

	/* libavcodec requires interleaved data.  It is simpler here to
	 * just queue interleaved samples to a single ringbuffer. */
	for (i = 0; i < nframes; i++) {
		for (chn = 0; chn < nports; chn++) {
			if (jack_ringbuffer_write (rb, (void *) (in[chn]+i),
					      sample_size)
			    < sample_size)
				overruns++;
		}
	}

	/* Tell the work thread there is work to do.  If it is already
	 * running, the lock will not be available.  We can't wait
	 * here in the process() thread, but we don't need to signal
	 * in that case, because the work thread will read all the
	 * data queued before waiting again. */
	if (pthread_mutex_trylock (&work_thread_lock) == 0) {
	    pthread_cond_signal (&data_ready);
	    pthread_mutex_unlock (&work_thread_lock);
	}

	return 0;
}

void
jack_shutdown (void *arg)
{
	fprintf (stderr, "JACK shutdown\n");
	/* exit (0); */
	abort();
}


void
setup_work_thread (thread_info_t *info)
{
	AVCodec * codec;
	int ret;
	
	avcodec_register_all();
	
	/* open ac3 codec */
	codec = avcodec_find_encoder(CODEC_ID_AC3);
	if (!codec) {
		fprintf(stderr, "AC3 codec not found\n");
		jack_client_close (info->client);
		exit(1);
	}
	
	info->av_context = avcodec_alloc_context();
	
	/* put sample parameters */
	info->av_context->bit_rate = info->bitrate * 1000;
	info->sample_rate = info->av_context->sample_rate = jack_get_sample_rate(info->client);
	info->av_context->channels = info->channels;
	
	/* open it */
	if ((ret = avcodec_open(info->av_context, codec)) < 0) {
		fprintf(stderr, "Error: could not open AC3 codec: invalid bitrate (%dkbps) or samplerate (%dHz)\n", info->av_context->bit_rate / 1000,info->sample_rate);
		jack_client_close (info->client);
		exit(1);
	}
	
	/* the codec gives us the frame size, in samples */
	info->av_frame_size = info->av_context->frame_size;
	
	info->duration *= info->av_context->sample_rate;
	info->can_capture = 0;

	info->dither_state = (dither_state_t *) malloc(sizeof(dither_state_t));

	/* open output file */
	info->sf = NULL;
	
	if (info->path) {
		if (strcmp(info->path, "stdout") == 0) {
			info->sf = stdout;
		}
		else {
			info->sf = fopen(info->path, "w+");
		}
	}

	
	pthread_create (&info->thread_id, NULL, work_thread, info);
}

void
run_work_thread (thread_info_t *info)
{
	info->can_capture = 1;
	pthread_join (info->thread_id, NULL);
	if (info->sf) {
		fclose (info->sf);
	}
	if (overruns > 0) {
		fprintf (stderr,
			 "ac3jack failed with %ld overruns.\n", overruns);
		fprintf (stderr, " try a bigger buffer than -B %"
			 PRIu32 ".\n", info->rb_size);
		info->status = EPIPE;
	}
	if (info->status) {
		/* unlink (info->path); */
	}
}

void
setup_ports (int sources, char *source_names[], thread_info_t *info)
{
	unsigned int i;
	size_t in_size;
	char portname[128];
	
	/* Allocate data structures that depend on the number of ports. */
	nports = info->channels;
	ports = (jack_port_t **) malloc (sizeof (jack_port_t *) * nports);
	in_size =  nports * sizeof (jack_default_audio_sample_t *);
	in = (jack_default_audio_sample_t **) malloc (in_size);
	rb = jack_ringbuffer_create (nports * sample_size * info->rb_size);

	/* When JACK is running realtime, jack_activate() will have
	 * called mlockall() to lock our pages into memory.  But, we
	 * still need to touch any newly allocated pages before
	 * process() starts using them.  Otherwise, a page fault could
	 * create a delay that would force JACK to shut us down. */
	memset(in, 0, in_size);
	memset(rb->buf, 0, rb->size);

	ports[0] = jack_port_register (info->client, "Left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput|JackPortIsTerminal, 0);

	/*
	 * 2 channels = L, R
	 * 3 channels = L, C, R
	 * 4 channels = L, R, LS, RS
	 * 5 ch       = L, C, R,  LS, RS
	 * 6 ch       = L, C, R,  LS, RS, LFE
	 */
	
	for (i=1; i < info->channels; ++i) {
		switch(i)
		{
		case 1:
			if (info->channels == 2 || info->channels == 4)
			{
				strncpy ( portname, "Right", sizeof(portname));
			}
			else {
				strncpy ( portname, "Center", sizeof(portname));
			}
		  
			break;
		case 2:
			if (info->channels == 4) {
				strncpy (portname, "LeftSurround", sizeof(portname) );
			} else {
				strncpy (portname, "Right",  sizeof(portname));
			}
			break;
		case 3:
			if (info->channels == 4) {
				strncpy (portname, "RightSurround", sizeof(portname));
			}
			else {
				strncpy (portname, "LeftSurround", sizeof(portname));
			}
			break;
		case 4:
			if (info->channels > 4) {
				strncpy (portname, "RightSurround", sizeof(portname));
			}
			break;
		case 5:
			strncpy (portname, "LFE", sizeof(portname));
			break;
		default:
			break;
		}
		
		
		ports[i] = jack_port_register (info->client, portname, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput|JackPortIsTerminal, 0);
	}
	
/* 	for (i = 0; i < nports; i++) { */
/* 		char name[64]; */

/* 		sprintf (name, "input%d", i+1); */

/* 		if ((ports[i] = jack_port_register (info->client, name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0)) == 0) { */
/* 			fprintf (stderr, "cannot register input port \"%s\"!\n", name); */
/* 			jack_client_close (info->client); */
/* 			exit (1); */
/* 		} */
/* 	} */

	for (i = 0; i < sources; i++) {
		if (i < nports && jack_connect (info->client, source_names[i], jack_port_name (ports[i]))) {
			fprintf (stderr, "cannot connect input port %s to %s\n", jack_port_name (ports[i]), source_names[i]);
		} 
	}

	info->can_process = 1;		/* process() can start, now */
}

void usage(char *argv0)
{
	fprintf(stderr, "ac3jack %s\nCopyright 2004 Jesse Chappell\nac3jack comes with ABSOLUTELY NO WARRANTY\n", ac3jack_version);
	fprintf(stderr, "This is free software, and you are welcome to redistribute it\n");
	fprintf(stderr, "under certain conditions; see the file COPYING for details\n\n");

	fprintf(stderr, "Usage: %s -o <str> [options...] [portname ...]\n", argv0);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -o <str> , --output=<str>    specify file to write output AC-3 stream, 'stdout' will write to standard output\n");
	fprintf(stderr, "  -c <num> , --channels=<num>  number of AC-3 output channels (1-6). default is 6 (5.1)\n");
	fprintf(stderr, "  -b <num> , --bitrate=<num>   bitrate of AC-3 stream (kbits/s). default is 448\n");
	fprintf(stderr, "  -d <num> , --duration=<num>  duration of output stream (secs). default is 0 (forever)\n");
	fprintf(stderr, "  -t , --jack-transport        use jack transport state (rolling/stopped) to control stream output\n");
	fprintf(stderr, "  -j <str> , --jack-name=<str> jack client name, default is ac3jack\n");
	fprintf(stderr, "  -B <num> , --bufsize=<num>   internal buffer size (samples), if overrun problems occur try a large value\n");
	fprintf(stderr, "  -q , --quiet                 do not output status to stderr\n");
	fprintf(stderr, "  -h , --help                  this usage output\n");
	fprintf(stderr, "  -V , --version               show version only\n");
	fprintf(stderr, "\nIf portname arguments are added, they are initially connected in channel order, if possible.\n");
}

int
main (int argc, char *argv[])

{
	jack_client_t *client;
	thread_info_t thread_info;
	int c;
	int longopt_index = 0;
	extern int optind, opterr;
	int show_usage = 0;
	int show_version = 0;
	char * jackname = NULL;
	char *optstring = "c:d:o:b:B:hqVtj:";
	struct option long_options[] = {
		{ "help", 0, 0, 'h' },
		{ "quiet", 0, 0, 'q' },
		{ "duration", 1, 0, 'd' },
		{ "channels", 1, 0, 'c' },
		{ "output", 1, 0, 'o' },
		{ "bitrate", 1, 0, 'b' },
		{ "bufsize", 1, 0, 'B' },
		{ "jack-name", 1, 0, 'j' },
		{ "jack-transport", 0, 0, 't' },
		{ "version", 0, 0, 'V' },
		{ 0, 0, 0, 0 }
	};

	memset (&thread_info, 0, sizeof (thread_info));
	thread_info.rb_size = DEFAULT_RB_SIZE;
	thread_info.path = NULL;
	thread_info.channels = 6;
	thread_info.bitrate = 448;
	thread_info.duration = 0;
	thread_info.use_transport = 0;
	thread_info.quiet = 0;
	opterr = 0;

	setupSignals();
	
	while ((c = getopt_long (argc, argv, optstring, long_options, &longopt_index)) != -1) {
		switch (c) {
		case 1:
			/* getopt signals end of '-' options */
			break;

		case 'h':
			show_usage++;
			break;
		case 'V':
			show_version++;
			break;
		case 't':
			thread_info.use_transport = 1;
			break;
		case 'q':
			thread_info.quiet = 1;
			break;
		case 'j':
			jackname = optarg;
			break;
		case 'd':
			thread_info.duration = atoi (optarg);
			break;
		case 'o':
			thread_info.path = optarg;
			break;
		case 'c':
			thread_info.channels = atoi(optarg);
			break;
		case 'b':
			thread_info.bitrate = atoi (optarg);
			break;
		case 'B':
			thread_info.rb_size = atoi (optarg);
			break;
		default:
			fprintf (stderr, "error\n");
			show_usage++;
			break;
		}
	}

	if (!thread_info.path) {
		show_usage++;
	}

	if (show_version) {
		fprintf(stderr, "ac3jack %s\nCopyright 2004 Jesse Chappell\n", ac3jack_version);
		exit(0);
	}

	if (show_usage ) {
		usage(argv[0]);
		exit (1);
	}
	
	if ((client = jack_client_new (jackname!=NULL ? jackname : "ac3jack")) == 0) {
		fprintf (stderr, "error connecting to jack server: invalid name or server not running.\n");
		exit (1);
	}

	thread_info.client = client;
	thread_info.can_process = 0;
	thread_info.do_write = 1;
	
	if (thread_info.channels < 1) {
		thread_info.channels = 1;
	}
	else if (thread_info.channels > 6) {
		thread_info.channels = 6;
	}
	
	setup_work_thread (&thread_info);

	jack_set_process_callback (client, process, &thread_info);
	jack_on_shutdown (client, jack_shutdown, &thread_info);

	if (jack_activate (client)) {
		fprintf (stderr, "error: cannot activate client");
		exit(1);
	}

	setup_ports (argc - optind, &argv[optind], &thread_info);

	if (!thread_info.quiet)
	{
		fprintf (stderr, "ac3jack encoding with %d channels, %dkHz, %d kbit/s bitrate",
			 thread_info.channels, thread_info.sample_rate/1000, thread_info.bitrate);
		if (thread_info.sf) {
			fprintf(stderr, ", output to %s", thread_info.path);
		}
		if (thread_info.duration) {
			fprintf(stderr, "  [%d seconds]", thread_info.duration / thread_info.sample_rate);
		}
		fprintf(stderr, "\n");

		if (thread_info.sf == stdout && thread_info.sample_rate != 48000) {
			fprintf(stderr, "WARNING: Not using 48kHz JACK sample rate.\nIf this output is being streamed to SPDIF (via ac3dec), there will be glitches!\n"); 
		}
	}
	
	run_work_thread (&thread_info);

	jack_client_close (client);

	jack_ringbuffer_free (rb);

	exit (0);
}
